import math
import unittest
import pyparaspace


def test_pyparaspace():
    problem = pyparaspace.Problem(
        timelines=[
            pyparaspace.Timeline(
                name="obj",
                token_types=[
                    pyparaspace.TokenType(
                        value="s1", conditions=[], duration_limits=(5, 6), capacity=0
                    ),
                    pyparaspace.TokenType(
                        value="s2",
                        conditions=[
                            pyparaspace.TemporalCond(
                                temporal_relation=pyparaspace.TemporalRelation.MetBy,
                                amount=0,
                                timeline="obj",
                                value="s1",
                            )
                        ],
                        duration_limits=(1, None),
                        capacity=0,
                    ),
                ],
                static_tokens=[
                    pyparaspace.StaticToken(
                        value="s2",
                        const_time=pyparaspace.goal(),
                        capacity=0,
                        conditions=[],
                    )
                ],
            )
        ],
    )

    solution = pyparaspace.solve(problem)
    print(f"Solution: {solution}")
    timeline = solution.timelines[0]
    assert len(timeline.tokens) == 2, "Number of tokens should be 2"

    token1 = timeline.tokens[0]
    token2 = timeline.tokens[1]
    assert token1.value == "s1", "Token value 1 should be s1"
    assert token2.value == "s2", "Token value 2 should be s2"

    assert (
        token1.end_time - token1.start_time >= 5.0
        and token1.end_time - token1.start_time <= 6.0
    )
    assert abs(token1.end_time - token2.start_time) < 1e-5


if __name__ == "__main__":
    test_pyparaspace()



from timelinedsl import *

for plates in [1,2]:
    for n_carbonaras in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20, 25, 30, 40, 50, 75, 100]:

        p = Problem()

        for _ in range(plates):
            p.resource("Plate", capacity=1)

        water1 = p.timeline("Water")
        water1.state("Heating", dur=(10, 10), conditions=[UseResource(Any("Plate"), 1)])
        water1.state("HotWater", conditions=[TransitionFrom("Heating")])

        oil1 = p.timeline("Oil")
        oil1.state("Heating", dur=(10, 10), conditions=[UseResource(Any("Plate"), 1)])
        oil1.state("HotOil", conditions=[TransitionFrom("Heating")])

        for i in range(n_carbonaras):
            spaghetti = p.timeline(classname="Spaghetti", name=f"spaghetti_{i}")
            spaghetti.state("Cooking", dur=(5, 5), conditions=[During(Any("Water"), "HotWater")])
            spaghetti.state("Cooked", conditions=[TransitionFrom("Cooking")])

            lardon = p.timeline(classname="Lardon", name=f"lardon_{i}")
            lardon.state("Cooking", dur=(5, 5), conditions=[During(Any("Oil"), "HotOil")])
            lardon.state("Cooked", conditions=[TransitionFrom("Cooking")])

            eggs = p.timeline(classname="Eggs", name=f"eggs_{i}")
            eggs.state("Beating", dur=(5, 5))
            eggs.state("Beaten", conditions=[TransitionFrom("Beating")])

            carbonara = p.timeline(classname="Carbonara", name=f"carbonara_{i}")
            carbonara.state("Cooking", dur=(3, 3), conditions=[
                StartsAfter(f"spaghetti_{i}", "Cooked"),
                StartsAfter(f"lardon_{i}", "Cooked"),
                StartsAfter(f"eggs_{i}", "Beaten"),
                UseResource(Any("Plate"), 1)
            ])
            carbonara.state("Cooked", conditions=[TransitionFrom("Cooking")])
            carbonara.state("Eating", dur=(5, 5), conditions=[TransitionFrom("Cooked")])
            carbonara.state("Eaten", conditions=[TransitionFrom("Eating")])

            p.goal(f"carbonara_{i}", "Eaten")

        p.save_json(f"examples/carbonara_{plates}p_{n_carbonaras}c.json")



import json
from collections import defaultdict, namedtuple
from dataclasses import dataclass
from typing import Any


class Problem():
    def __init__(self):
        self.timelines = []
        self.tokens = []
        self.groups = defaultdict(list)

    def resource(self, classname, name=None, capacity=0):
        if name is None:
            name = f"{classname}_{len(self.tokens)}"
        self.tokens.append({ "timeline_name": name, "value": "Available",  "const_time": { "Fact": [None, None] }, "capacity": capacity, "conditions": []})
        self.groups[classname].append(name)

    def timeline(self, classname, name=None):
        if name is None:
            name = f"{classname}_{len(self.timelines)}"
        timeline = Timeline(classname, name)
        self.timelines.append(timeline)
        self.groups[classname].append(name)
        return timeline

    def goal(self, timeline, value, capacity=0):
        self.tokens.append({"timeline_name": timeline, "value": value, "const_time": { "Goal": None }, "capacity": capacity, "conditions": []})

    def fact(self, timeline, value, start=None, end=None, capacity=0):
        self.tokens.append({"timeline_name": timeline, "value": value, "const_time": { "Fact": [start,end] }, "capacity": capacity, "conditions": []})

    def to_dict(self):
        return {"groups": [{"name": key, "members": value } for key,value in self.groups.items()],
                "timelines": list(map(lambda t: t.to_dict(), self.timelines)),
                "tokens": self.tokens }

    def save_json(self,fn):
        with open(fn,"w") as f:
            json.dump(self.to_dict(), f, indent=2)
        print(f"Wrote problem instance to file {fn}")


@dataclass
class UseResource:
    resource :Any
    amount :int

@dataclass
class TransitionFrom:
    value :Any

@dataclass
class TransitionTo:
    value :Any

@dataclass
class MetBy:
    timelineref: Any
    value :Any
    amount :int = 0

@dataclass
class Meets:
    timelineref: Any
    value :Any
    amount :int = 0

@dataclass
class StartsAfter:
    timelineref: Any
    value :Any
    amount :int = 0

@dataclass
class During:
    timelineref :Any
    value :Any
    amount :int = 0

@dataclass
class Any:
    classname :Any

# UseResource = namedtuple("UseResource", "resource,amount")
# TransitionFrom = namedtuple("TransitionFrom", "value")
# MetBy = namedtuple("MetBy", "timelineref,value,amount")
# During = namedtuple("During", "timelineref,value,amount")
# Any = namedtuple("Any", "classname")

class Timeline():
    def __init__(self, classname, name=None):
        self.classname = classname
        self.name = name
        self.states = []

    def state(self, name, dur=(1,None), capacity=0, conditions=None):
        self.states.append({"name": name, "duration": dur, "capacity": capacity,
                            "conditions": list(map(lambda cond: condition_to_dict(self.name, cond), conditions or []))})

    def to_dict(self):
        return {
            "name": self.name,
            "values": self.states,
        }

def objectref_to_dict(objectref):
    if isinstance(objectref, Any):
        return {"Group": objectref.classname}
    else:
        return {"Object": objectref}

def condition_to_dict(obj_name, condition):
    if isinstance(condition, UseResource):
        return { "temporal_relationship": "Cover", "object": objectref_to_dict(condition.resource), "value": "Available", "amount": condition.amount }
    elif isinstance(condition, TransitionFrom):
        return { "temporal_relationship": "MetBy", "object": objectref_to_dict(obj_name), "value": condition.value, "amount": 0}
    elif isinstance(condition, TransitionTo):
        return { "temporal_relationship": "Meets", "object": objectref_to_dict(obj_name), "value": condition.value, "amount": 0}
    elif isinstance(condition, StartsAfter):
        return { "temporal_relationship": "StartsAfter", "object": objectref_to_dict(condition.timelineref), "value": condition.value, "amount": 0}
    elif isinstance(condition, MetBy):
        return { "temporal_relationship": "MetBy", "object":  objectref_to_dict(condition.timelineref), "value": condition.value, "amount": condition.amount }
    elif isinstance(condition, Meets):
        return { "temporal_relationship": "Meets", "object":  objectref_to_dict(condition.timelineref), "value": condition.value, "amount": condition.amount }
    elif isinstance(condition, During):
        return { "temporal_relationship": "Cover", "object":  objectref_to_dict(condition.timelineref), "value": condition.value, "amount": condition.amount }
    raise Exception(f"Unknown condition type {condition}")



import sys, json
import matplotlib
import matplotlib.pyplot as plt
from collections import defaultdict

colors = ["red","green","blue","purple","orange","black","gray"]

filename = sys.argv[1]
print(f"Reading file {filename}")

with open(filename, "r") as f:
    model = json.load(f)

print(f"got model {model}")

fig, ax = plt.subplots(figsize=(10,6),dpi=120)

def rat(x):
    return float(x[0]) / float(x[1])

y_stack = []
y_names = {}

timelines = defaultdict(list)
for token in model["tokens"]:
    timelines[token["object_name"]].append(token)

for (timeline_name,tokens) in timelines.items():
    y_names[timeline_name] = len(y_stack)
    intervals = []
    for token in tokens:
        start = token["start_time"] if token["start_time"] is not None else  token["end_time"] - 1
        end = token["end_time"] if token["end_time"] is not None else  token["start_time"] + 1
        intervals.append((start, end, colors[hash(token["value"]) % len(colors)], token["value"]))
    y_stack.append( { "name": timeline_name, "intervals" : intervals })

#y_names["visibility"] = len(y_stack)
#y_stack.append({ "name": "visibility", "intervals": [(rat(x[0]),rat(x[1]),"red") for x in model["problem"]["visibility_time_windows"]] })
#
#for (l_idx, location) in enumerate(model["locations"]):
#    name = f"loc{l_idx}_@{location[0]}"
#    y_names[name] = len(y_stack)
#    y_stack.append({"name": name, "intervals": [(location[1][0], location[1][1], "blue")]})
#
#for (_idx, take_picture) in enumerate(model["take_picture"]):
#    name = f"take_@{take_picture[0]}"
#    y_names[name] = len(y_stack)
#    y_stack.append({"name": name, "intervals": [(take_picture[1][0], take_picture[1][1], "green")]})
#
#for (_idx, download) in enumerate(model["download"]):
#    name = f"dl_{download[0]}"
#    y_names[name] = len(y_stack)
#    y_stack.append({"name": name, "intervals": [(download[1][0], download[1][1], "orange")]})

#for (idx,water) in enumerate(model["problem"]["visibility_time_windows"]):
#    name = f"water{idx}"
#for (idx,oil) in enumerate(model["oil_heat"]):
#    name = f"oil{idx}"
#    y_names[name] = len(y_stack)
#    y_stack.append({ "name": name, "intervals": [(oil[1],oil[2],"red")] })
#
#for (idx,carbonara) in enumerate(model["carbonaras"]):
#    name = f"carbonara{idx}"
#    y_names[name] = len(y_stack)
#    y_stack.append({ "name": name, "intervals": [
#        (carbonara["spaghetti_start"],carbonara["spaghetti_end"],"red"),
#        (carbonara["lardon_start"],carbonara["lardon_end"],"green"),
#        (carbonara["eggs_start"],carbonara["eggs_end"],"blue"),
#        (carbonara["cook_start"],carbonara["cook_end"],"orange"),
#        (carbonara["eat_start"],carbonara["eat_end"],"purple"),
#    ]})


ax.set_ylim([-0.5, len(y_stack)-1 +0.5])
ax.set_yticks(list(range(len(y_stack))))
ax.set_yticklabels([x["name"] for x in y_stack])
fig.subplots_adjust(left=0.3)

for item in y_stack:
    y = y_names[item["name"]]
    print(f"item {item}")
    for i,x in enumerate(item["intervals"]):
        ax.broken_barh([(x[0],x[1]-x[0])], 
                       (y-0.4+0.1*i,0.1),
                       facecolor=[x[2]])
        ax.text(x[0], y-0.4+0.1*i+0.05, x[3])

# ARROWS
#for (structure_idx,structure) in enumerate(model["problem"]["structures"]):
#    for p_idx in [structure["piece_1"], structure["piece_2"]]:
#        p_x = model["pieces"][p_idx]["treated"]
#        p_y = y_names[f"piece{p_idx}"]
#        s_x = model["structures"][structure_idx]["assemble"]
#        s_y = y_names[f"structure{structure_idx}"]
#        ax.arrow(p_x, p_y, s_x - p_x, s_y - p_y, color="blue", head_width=0.1, head_length=0.15, length_includes_head=True)


plt.savefig(f"{filename}.plot.png")



from timelinedsl import *

time_windows = [(280,4000), (8000,15000), (20000,25000), (30000,35000), (40000,45000)]
locations = [0, 1,2,3,4,5,6,7,8,9]
speed = 1
download_time = 6*40
take_pic_time = 50
edges = [
        (0, 1, 100),
        (0, 2, 200),
        (0, 4, 100),
        (0, 5, 200),
        (0, 9, 300),
        (1, 2, 100),
        (1, 6, 200),
        (2, 3, 100),
        (2, 7, 200),
        (2, 8, 300),
        (3, 4, 200),
        (3, 7, 100),
        (3, 8, 200),
        (4, 5, 100),
        (4, 9, 200),
        (5, 6, 100),
        (5, 7, 200),
        (5, 9, 100),
        (6, 7, 100),
        (7, 8, 100),
        (8, 9, 200),
]

edges = edges + [(b,a,l) for a,b,l in edges]


for n_pics in [1,2,3,4,5,6,7,8,9]:
    for n_windows in [1,2,3,4,5]:

        p = Problem()
        p.resource("Antenna", name="Antenna", capacity=1)
        for (start,end) in time_windows[:n_windows]:
            p.fact("Visibility", "Available", start=start, end=end)
        
        loc_timeline = p.timeline("Location", name="loc")
        p.fact("loc", "At(0)",start=0) # Start at location 0

        for loc in locations:
            loc_timeline.state(f"At({loc})")
        for a,b,l in edges:
            loc_timeline.state(f"Going({locations[a]},{locations[b]})", dur=(l,l), conditions=[
                TransitionFrom(f"At({locations[a]})"), 
                TransitionTo(f"At({locations[b]})")])
        
        for loc_idx in range(n_pics):

            # Take the picture
            pic_timeline = p.timeline("HavePicture", name=f"HavePicture{loc_idx}")
            pic_timeline.state("Taking", dur=(take_pic_time, take_pic_time),conditions=[During("loc", f"At({loc_idx})")])
            pic_timeline.state("Done", conditions=[TransitionFrom("Taking")])

            # Download the picture
            dl_timeline = p.timeline("Download", name=f"Download{loc_idx}")
            dl_timeline.state("Downloading", dur=(download_time, download_time), conditions=[
                During(f"HavePicture{loc_idx}", "Done"), 
                During(f"Visibility", "Available"),
                During(f"Antenna", "Available", amount=1)])
            dl_timeline.state("Done", conditions=[TransitionFrom("Downloading")])

            p.goal(f"Download{loc_idx}", "Done")

        p.save_json(f"examples/goac_{n_pics}pics_{n_windows}wind.json")



from timelinedsl import *

for (n_kilns,n_pieces) in [(1,2),(2,4),(2,6),(4,6),(5,10),(1,10)]:

    p = Problem()
    p.resource("Electricity", capacity=1)

    for kiln_idx in range(n_kilns):
        kiln = p.timeline("Kiln", f"kiln_{kiln_idx}")
        kiln.state("Ready", dur=(5,None), conditions=[TransitionFrom("Fire")])
        kiln.state("Fire", dur=(20,20), capacity=2, conditions=[
            TransitionFrom("Ready"), 
            UseResource(Any("Electricity"), 1)])
        p.fact(f"kiln_{kiln_idx}", "Ready")

    piece_param_types = [(5,2),(8,3),(11,1)]
    pieces = []
    while len(pieces) < n_pieces:
        pieces.append(piece_param_types[len(pieces) % len(piece_param_types)])

    for (piece_idx,(bake_time,treat_time)) in enumerate(pieces):
        piece = p.timeline("Piece", f"piece_{piece_idx}")
        piece.state("Baking", dur=(bake_time,bake_time), conditions=[
            During(Any("Kiln"), "Fire", 1),
        ])
        piece.state("Baked", conditions=[TransitionFrom("Baking")])
        piece.state("Treating", dur=(treat_time, treat_time), 
            conditions=[TransitionFrom("Baked")])
        piece.state("Treated", conditions=[TransitionFrom("Treating")])

        if piece_idx >= 2*(n_pieces//2):
            p.goal(f"piece_{piece_idx}", "Baked")

    for structure_idx in range(n_pieces // 2):
        structure = p.timeline("Structure", f"structure_{structure_idx}")
        structure.state("Assembling", dur=(1,1), conditions=[
            During(f"piece_{2*structure_idx}", "Treated"),
            During(f"piece_{2*structure_idx +1}", "Treated")
        ])
        structure.state("Assembled", conditions=[TransitionFrom("Assembling")])
        structure.state("Baking", dur=(3,3), conditions=[
            During(Any("Kiln"), "Fire", 1),
            TransitionFrom("Assembled"),
        ])
        structure.state("Baked", conditions=[TransitionFrom("Baking")])
        p.goal(f"structure_{structure_idx}", "Baked")

    p.save_json(f"examples/ceramic_{n_kilns}m_{n_pieces}j.json")
